/*
 * Copyright (C) 1997-2007 Andre Noll <maan@systemlinux.org>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */

/** \file client.c the client program used to connect to para_server */

#include "para.h"
#include "list.h"
#include "sched.h"
#include "client.cmdline.h"
#include "string.h"
#include "stdin.h"
#include "stdout.h"
#include "client.h"
#include "error.h"

INIT_CLIENT_ERRLISTS;

static struct private_client_data *pcd;
static struct stdin_task sit;
static struct stdout_task sot;


INIT_STDERR_LOGGING(pcd->conf.loglevel_arg);

static void client_event_handler(struct task *t)
{
	struct private_client_data *p = t->private_data;

	PARA_NOTICE_LOG("%s\n", PARA_STRERROR(-t->ret));
	if (t->ret != -E_HANDSHAKE_COMPLETE) {
		unregister_task(t);
		p->eof = 1;
		return;
	}
	if (p->status == CL_SENDING) {
		stdin_set_defaults(&sit);
		sit.buf = para_malloc(sit.bufsize),
		register_task(&sit.task);
		p->inbuf = sit.buf;
		p->in_loaded = &sit.loaded;
		p->in_eof = &sit.eof;
		return;
	}
	stdout_set_defaults(&sot);
	sot.buf = p->buf;
	sot.loaded = &p->loaded;
	sot.input_eof = &p->eof;
	register_task(&sot.task);
}

/**
 * the client program to connect to para_server
 *
 * \param argc usual argument count
 * \param argv usual argument vector
 *
 * It registers two tasks: The client task that communicates with para_server
 * and the standard out task that writes any output produced by the client task
 * to standard out.
 *
 * \return EXIT_SUCCESS or EXIT_FAILURE
 *
 * \sa client_open(), stdout.c, stdout.h, para_client(1), para_server(1)
 */
int main(int argc, char *argv[])
{

	int ret;
	struct sched s;

	s.default_timeout.tv_sec = 1;
	s.default_timeout.tv_usec = 0;
	ret = client_open(argc, argv, &pcd);
	if (ret < 0) /* can not use PARA_LOG here */
		exit(EXIT_FAILURE);
	pcd->task.event_handler = client_event_handler;
	ret = sched(&s);
	if (ret < 0)
		PARA_ERROR_LOG("%s\n", PARA_STRERROR(-ret));
	client_close(pcd);
	return ret >= 0? EXIT_SUCCESS: EXIT_FAILURE;
}
